"""nile common module."""
import logging
import os

from nile.common import DECLARATIONS_FILENAME, DEPLOYMENTS_FILENAME


def register(address, abi, network, alias):
    """Register a new deployment."""
    file = f"{network}.{DEPLOYMENTS_FILENAME}"

    if alias is not None:
        if exists(alias, network):
            raise Exception(f"Alias {alias} already exists in {file}")

    with open(file, "a") as fp:
        if alias is not None:
            logging.info(f"📦 Registering deployment as {alias} in {file}")
        else:
            logging.info(f"📦 Registering {address} in {file}")

        fp.write(f"{address}:{abi}")
        if alias is not None:
            fp.write(f":{alias}")
        fp.write("\n")


def register_class_hash(hash, network, alias):
    """Register a new deployment."""
    file = f"{network}.{DECLARATIONS_FILENAME}"

    if class_hash_exists(hash, network):
        raise Exception(f"Hash {hash[:6]}...{hash[-6:]} already exists in {file}")

    with open(file, "a") as fp:
        if alias is not None:
            logging.info(f"📦 Registering {alias} in {file}")
        else:
            logging.info(f"📦 Registering {hash} in {file}")

        fp.write(f"{hash}")
        if alias is not None:
            fp.write(f":{alias}")
        fp.write("\n")


def exists(identifier, network):
    """Return whether a deployment exists or not."""
    foo = next(load(identifier, network), None)
    return foo is not None


def class_hash_exists(hash, network):
    """Return whether a class declaration exists or not."""
    if hash in load_class(hash, network):
        return True


def load(identifier, network):
    """Load deployments that matches an identifier (address or alias)."""
    file = f"{network}.{DEPLOYMENTS_FILENAME}"

    if not os.path.exists(file):
        return

    with open(file) as fp:
        for line in fp:
            [address, abi, *alias] = line.strip().split(":")
            identifiers = [x for x in [address] + alias]
            if identifier in identifiers:
                yield address, abi


def load_class(identifier, network):
    """Load declaration class that matches an identifier (hash or alias)."""
    file = f"{network}.{DECLARATIONS_FILENAME}"

    if not os.path.exists(file):
        return

    with open(file) as fp:
        for line in fp:
            [hash, *alias] = line.strip().split(":")
            identifiers = [x for x in [hash] + alias]
            if identifier in identifiers:
                yield hash



"""Command to declare StarkNet smart contracts."""
import logging

from nile import deployments
from nile.common import DECLARATIONS_FILENAME, parse_information, run_command


def declare(contract_name, network, alias=None, overriding_path=None):
    """Declare StarkNet smart contracts."""
    logging.info(f"🚀 Declaring {contract_name}")

    if alias_exists(alias, network):
        file = f"{network}.{DECLARATIONS_FILENAME}"
        raise Exception(f"Alias {alias} already exists in {file}")

    output = run_command(contract_name, network, overriding_path, operation="declare")
    class_hash, tx_hash = parse_information(output)
    logging.info(f"⏳ Declaration of {contract_name} successfully sent at {class_hash}")
    logging.info(f"🧾 Transaction hash: {tx_hash}")

    deployments.register_class_hash(class_hash, network, alias)
    return class_hash


def alias_exists(alias, network):
    """Return whether an alias exists or not."""
    existing_alias = next(deployments.load_class(alias, network), None)
    return existing_alias is not None



"""
Tests for compile command.

Only unit tests for now. No contracts are actually compiled.
"""

import logging
from unittest.mock import Mock, patch

import pytest

from nile.common import ABIS_DIRECTORY, CONTRACTS_DIRECTORY
from nile.core.compile import _compile_contract, compile

CONTRACT = "foo.cairo"


@pytest.fixture(autouse=True)
def tmp_working_dir(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    return tmp_path


# Mock `subprocess` for all tests in this module to avoid actually running any
# shell commands.
@pytest.fixture(autouse=True)
def mock_subprocess():
    with patch("nile.core.compile.subprocess") as mock_subprocess:
        yield mock_subprocess


def test_compile_create_abis_directory(tmp_working_dir):
    assert not (tmp_working_dir / ABIS_DIRECTORY).exists()
    compile([])
    assert (tmp_working_dir / ABIS_DIRECTORY).exists()


@patch("nile.core.compile.get_all_contracts")
def test_compile_get_all_contracts_called(mock_get_all_contracts):
    compile([])
    mock_get_all_contracts.assert_called_once()
    mock_get_all_contracts.reset_mock()

    compile([CONTRACT])
    mock_get_all_contracts.assert_not_called()


@patch("nile.core.compile._compile_contract")
def test_compile__compile_contract_called(mock__compile_contract):
    compile([CONTRACT])
    mock__compile_contract.assert_called_once_with(
        CONTRACT, CONTRACTS_DIRECTORY, False, False
    )


@patch("nile.core.compile._compile_contract")
def test_compile_failure_feedback(mock__compile_contract, caplog):
    # make logs visible to test
    logging.getLogger().setLevel(logging.INFO)

    mock__compile_contract.side_effect = [0]
    compile([CONTRACT])
    assert "Done" in caplog.text

    mock__compile_contract.side_effect = [1]
    compile([CONTRACT])
    assert "Failed" in caplog.text


def test__compile_contract(mock_subprocess):
    contract_name_root = "contract"
    path = f"path/to/{contract_name_root}.cairo"

    mock_process = Mock()
    mock_subprocess.Popen.return_value = mock_process

    _compile_contract(path)

    mock_subprocess.Popen.assert_called_once_with(
        [
            "starknet-compile",
            path,
            f"--cairo_path={CONTRACTS_DIRECTORY}",
            "--output",
            f"artifacts/{contract_name_root}.json",
            "--abi",
            f"artifacts/abis/{contract_name_root}.json",
        ],
        stdout=mock_subprocess.PIPE,
    )
    mock_process.communicate.assert_called_once()


def test__compile_account_contract(mock_subprocess):
    contract_name_root = "mock_account"
    path = f"path/to/{contract_name_root}.cairo"

    mock_process = Mock()
    mock_subprocess.Popen.return_value = mock_process

    _compile_contract(path, account_contract="--account_contract")

    mock_subprocess.Popen.assert_called_once_with(
        [
            "starknet-compile",
            path,
            f"--cairo_path={CONTRACTS_DIRECTORY}",
            "--output",
            f"artifacts/{contract_name_root}.json",
            "--abi",
            f"artifacts/abis/{contract_name_root}.json",
            "--account_contract",
        ],
        stdout=mock_subprocess.PIPE,
    )
    mock_process.communicate.assert_called_once()


@pytest.mark.parametrize(
    "contract_name, flag",
    [
        ("Account", "--account_contract"),
        ("mock_Account", "--account_contract"),
        ("Account_Manager", False),
    ],
)
def test__compile_auto_account_flag(mock_subprocess, contract_name, flag):
    path = f"path/to/{contract_name}.cairo"

    mock_process = Mock()
    mock_subprocess.Popen.return_value = mock_process

    _compile_contract(path)

    returned_subprocess = [
        "starknet-compile",
        path,
        f"--cairo_path={CONTRACTS_DIRECTORY}",
        "--output",
        f"artifacts/{contract_name}.json",
        "--abi",
        f"artifacts/abis/{contract_name}.json",
    ]

    if flag is not False:
        returned_subprocess.append(flag)

    mock_subprocess.Popen.assert_called_once_with(
        returned_subprocess,
        stdout=mock_subprocess.PIPE,
    )
    mock_process.communicate.assert_called_once()


def test__compile_contract_without_hint_validation(mock_subprocess):
    contract_name_root = "contract_with_unwhitelisted_hints"
    path = f"path/to/{contract_name_root}.cairo"

    mock_process = Mock()
    mock_subprocess.Popen.return_value = mock_process

    _compile_contract(path, disable_hint_validation=True)

    mock_subprocess.Popen.assert_called_once_with(
        [
            "starknet-compile",
            path,
            f"--cairo_path={CONTRACTS_DIRECTORY}",
            "--output",
            f"artifacts/{contract_name_root}.json",
            "--abi",
            f"artifacts/abis/{contract_name_root}.json",
            "--disable_hint_validation",
        ],
        stdout=mock_subprocess.PIPE,
    )
    mock_process.communicate.assert_called_once()



"""Tests for deploy command."""
import logging
from unittest.mock import patch

import pytest

from nile.core.deploy import ABIS_DIRECTORY, BUILD_DIRECTORY, deploy


@pytest.fixture(autouse=True)
def tmp_working_dir(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    return tmp_path


CONTRACT = "contract"
NETWORK = "goerli"
ALIAS = "alias"
ABI = f"{ABIS_DIRECTORY}/{CONTRACT}.json"
BASE_PATH = (BUILD_DIRECTORY, ABIS_DIRECTORY)
PATH_OVERRIDE = ("artifacts2", ABIS_DIRECTORY)
RUN_OUTPUT = b"output"
ARGS = [1, 2, 3]
ADDRESS = 999
TX_HASH = 222


@pytest.mark.parametrize(
    "args, exp_command",
    [
        (
            [CONTRACT, ARGS, NETWORK, ALIAS],  # args
            [CONTRACT, NETWORK, None],  # expected command
        ),
        (
            [CONTRACT, ARGS, NETWORK, ALIAS, PATH_OVERRIDE],  # args
            [CONTRACT, NETWORK, PATH_OVERRIDE],  # expected command
        ),
    ],
)
@patch("nile.core.deploy.run_command", return_value=RUN_OUTPUT)
@patch("nile.core.deploy.parse_information", return_value=[ADDRESS, TX_HASH])
@patch("nile.core.deploy.deployments.register")
def test_deploy(mock_register, mock_parse, mock_run_cmd, caplog, args, exp_command):
    logging.getLogger().setLevel(logging.INFO)

    # check return values
    res = deploy(*args)
    assert res == (ADDRESS, ABI)

    # check internals
    mock_run_cmd.assert_called_once_with(*exp_command, arguments=ARGS)
    mock_parse.assert_called_once_with(RUN_OUTPUT)
    mock_register.assert_called_once_with(ADDRESS, ABI, NETWORK, ALIAS)

    # check logs
    assert f"🚀 Deploying {CONTRACT}" in caplog.text
    assert f"⏳ ️Deployment of {CONTRACT} successfully sent at {ADDRESS}" in caplog.text
    assert f"🧾 Transaction hash: {TX_HASH}" in caplog.text



"""Tests for account commands."""
from unittest.mock import MagicMock, patch

import pytest

from nile.core.account import Account

KEY = "TEST_KEY"
NETWORK = "goerli"
MOCK_ADDRESS = "0x123"
MOCK_INDEX = 0


@pytest.fixture(autouse=True)
def tmp_working_dir(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    return tmp_path


@patch("nile.core.account.Account.deploy")
def test_account_init(mock_deploy):
    mock_deploy.return_value = MOCK_ADDRESS, MOCK_INDEX
    account = Account(KEY, NETWORK)

    assert account.address == MOCK_ADDRESS
    assert account.index == MOCK_INDEX
    mock_deploy.assert_called_once()


def test_account_multiple_inits_with_same_key():
    account = Account(KEY, NETWORK)
    account.deploy()
    account2 = Account(KEY, NETWORK)

    # Check addresses don't match
    assert account.address != account2.address
    # Check indexing
    assert account.index == 0
    assert account2.index == 1


@patch("nile.core.account.deploy", return_value=(1, 2))
def test_deploy(mock_deploy):
    account = Account(KEY, NETWORK)
    with patch("nile.core.account.os.path.dirname") as mock_path:
        test_path = "/overriding_path"
        mock_path.return_value.replace.return_value = test_path

        account.deploy()

        mock_deploy.assert_called_with(
            "Account",
            [str(account.signer.public_key)],
            NETWORK,
            f"account-{account.index + 1}",
            (f"{test_path}/artifacts", f"{test_path}/artifacts/abis"),
        )


@patch("nile.core.account.deploy", return_value=(MOCK_ADDRESS, MOCK_INDEX))
@patch("nile.core.account.accounts.register")
def test_deploy_accounts_register(mock_register, mock_deploy):
    account = Account(KEY, NETWORK)

    mock_register.assert_called_once_with(
        account.signer.public_key, MOCK_ADDRESS, MOCK_INDEX, NETWORK
    )


@patch("nile.core.account.call_or_invoke")
def test_send_nonce_call(mock_call):
    account = Account(KEY, NETWORK)
    contract_address, _ = account.deploy()

    # Instead of creating and populating a tmp .txt file, this uses the
    # deployed account address (contract_address) as the target
    account.send(contract_address, "method", [1, 2, 3], max_fee=1)

    # 'call_or_invoke' is called twice ('get_nonce' and '__execute__')
    assert mock_call.call_count == 2

    # Check 'get_nonce' call
    mock_call.assert_any_call(account.address, "call", "get_nonce", [], NETWORK)


@pytest.mark.parametrize(
    "callarray, calldata",
    # The following callarray and calldata args tests the Account's list comprehensions
    # ensuring they're set to strings and passed correctly
    [([[111]], []), ([[111, 222]], [333, 444, 555])],
)
def test_send_sign_transaction_and_execute(callarray, calldata):
    account = Account(KEY, NETWORK)
    contract_address, _ = account.deploy()

    sig_r, sig_s = [999, 888]
    return_signature = [callarray, calldata, sig_r, sig_s]

    account.signer.sign_transaction = MagicMock(return_value=return_signature)

    with patch("nile.core.account.call_or_invoke") as mock_call:
        send_args = [contract_address, "method", [1, 2, 3]]
        nonce = 4
        max_fee = 1
        account.send(*send_args, max_fee, nonce)

        # Check values are correctly passed to 'sign_transaction'
        account.signer.sign_transaction.assert_called_once_with(
            calls=[send_args], nonce=nonce, sender=account.address, max_fee=1
        )

        # Check values are correctly passed to '__execute__'
        mock_call.assert_called_with(
            contract=account.address,
            max_fee=str(max_fee),
            method="__execute__",
            network=NETWORK,
            params=[
                str(len(callarray)),
                *(str(elem) for sublist in callarray for elem in sublist),
                str(len(calldata)),
                *(str(param) for param in calldata),
                str(nonce),
            ],
            signature=[str(sig_r), str(sig_s)],
            type="invoke",
        )



"""nile common module."""
import json
import os

from nile.common import ACCOUNTS_FILENAME


def register(pubkey, address, index, network):
    """Register a new account."""
    file = f"{network}.{ACCOUNTS_FILENAME}"

    if exists(pubkey, network):
        raise Exception(f"account-{index} already exists in {file}")

    with open(file, "r") as fp:
        accounts = json.load(fp)
        accounts[pubkey] = {"address": address, "index": index}
    with open(file, "w") as file:
        json.dump(accounts, file)


def exists(pubkey, network):
    """Return whether an account exists or not."""
    account = next(load(pubkey, network), None)
    return account is not None


def load(pubkey, network):
    """Load account that matches a pubkey."""
    file = f"{network}.{ACCOUNTS_FILENAME}"

    if not os.path.exists(file):
        with open(file, "w") as fp:
            json.dump({}, fp)

    with open(file) as fp:
        accounts = json.load(fp)
        if pubkey in accounts:
            yield accounts[pubkey]


def current_index(network):
    """Return the length of the accounts. Used as the next index."""
    file = f"{network}.{ACCOUNTS_FILENAME}"

    with open(file) as fp:
        accounts = json.load(fp)
        return len(accounts.keys())



"""Utility for signing transactions for an Account on Starknet."""

from starkware.crypto.signature.signature import private_to_stark_key, sign
from starkware.starknet.core.os.transaction_hash.transaction_hash import (
    TransactionHashPrefix,
    calculate_transaction_hash_common,
)
from starkware.starknet.definitions.general_config import StarknetChainId
from starkware.starknet.public.abi import get_selector_from_name

TRANSACTION_VERSION = 0


class Signer:
    """Utility for signing transactions for an Account on Starknet."""

    def __init__(self, private_key):
        """Construct a Signer object. Takes a private key."""
        self.private_key = private_key
        self.public_key = private_to_stark_key(private_key)

    def sign(self, message_hash):
        """Sign a message hash."""
        return sign(msg_hash=message_hash, priv_key=self.private_key)

    def sign_transaction(self, sender, calls, nonce, max_fee):
        """Sign a transaction for an Account."""
        (call_array, calldata) = from_call_to_call_array(calls)
        message_hash = get_transaction_hash(
            int(sender, 16), call_array, calldata, nonce, int(max_fee)
        )
        sig_r, sig_s = self.sign(message_hash)
        return (call_array, calldata, sig_r, sig_s)


# Auxiliary functions


def from_call_to_call_array(calls):
    """Transform from Call to CallArray."""
    call_array = []
    calldata = []
    for _, call in enumerate(calls):
        assert len(call) == 3, "Invalid call parameters"
        entry = (
            int(call[0], 16),
            get_selector_from_name(call[1]),
            len(calldata),
            len(call[2]),
        )
        call_array.append(entry)
        calldata.extend(call[2])
    return (call_array, calldata)


def get_transaction_hash(account, call_array, calldata, nonce, max_fee):
    """Calculate the transaction hash."""
    execute_calldata = [
        len(call_array),
        *[x for t in call_array for x in t],
        len(calldata),
        *calldata,
        nonce,
    ]

    return calculate_transaction_hash_common(
        TransactionHashPrefix.INVOKE,
        TRANSACTION_VERSION,
        account,
        get_selector_from_name("__execute__"),
        execute_calldata,
        max_fee,
        StarknetChainId.TESTNET.value,
        [],
    )



"""Tests for debug command."""
import logging
import sys
from pathlib import Path
from unittest.mock import patch

import pytest

from nile.common import BUILD_DIRECTORY
from nile.utils.debug import _abi_to_build_path, _locate_error_lines_with_abis, debug

MOCK_HASH = "0x1234"
NETWORK = "goerli"
ERROR_MESSAGE = "Error at pc=0:1:\nAn ASSERT_EQ instruction failed: 3 != 0."
DEBUG_ADDRESS = "0x07826b88e404632d9835ab1ec2076c6cf1910e6ecb2ed270647fc211ff55e76f"
ABI_PATH = "path/to/abis/test_contract.json"
ALIAS = "contract_alias"
MOCK_FILE = 123


def mocked_json_message(arg):
    return {"tx_status": arg, "tx_failure_reason": {"error_message": ERROR_MESSAGE}}


@pytest.fixture(autouse=True)
def tmp_working_dir(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    return tmp_path


def test__abi_to_build_path():
    Path(BUILD_DIRECTORY).mkdir()
    filename = "contract"
    assert f"{BUILD_DIRECTORY}/{filename}" == _abi_to_build_path(filename)


@pytest.mark.parametrize(
    "file, address_set",
    [
        ([f"{DEBUG_ADDRESS}:{ABI_PATH}"], [int(DEBUG_ADDRESS, 16)]),
        ([f"{DEBUG_ADDRESS}:{ABI_PATH}:{ALIAS}"], [int(DEBUG_ADDRESS, 16)]),
    ],
)
@patch("nile.utils.debug._abi_to_build_path", return_value=ABI_PATH)
def test__locate_error_lines_with_abis_with_and_without_alias(
    mock_path, file, address_set
):
    with patch("nile.utils.debug.open") as mock_open:
        mock_open.return_value.__enter__.return_value = file
        return_array = _locate_error_lines_with_abis(MOCK_FILE, address_set, mock_path)
        # Values should be pushed into the array (and not return an error)
        # regardless if a contract alias is present or not
        assert return_array == [f"{DEBUG_ADDRESS}:{ABI_PATH}"]


@patch("nile.utils.debug._abi_to_build_path", return_value=ABI_PATH)
def test__locate_error_lines_with_abis_misformatted_line(mock_path, caplog):
    logging.getLogger().setLevel(logging.INFO)

    with patch("nile.utils.debug.open") as mock_open:
        # The DEBUG_ADDRESS alone without ":" is misformatted
        mock_open.return_value.__enter__.return_value = [DEBUG_ADDRESS]
        _locate_error_lines_with_abis(MOCK_FILE, int(DEBUG_ADDRESS, 16), mock_path)
        assert f"⚠ Skipping misformatted line #1 in {MOCK_FILE}" in caplog.text


@pytest.mark.parametrize(
    "args, expected",
    [
        ("ACCEPTED", "No error in transaction"),
        ("REJECTED", "The transaction was rejected"),
        ("REJECTED", ERROR_MESSAGE),
    ],
)
@pytest.mark.xfail(
    sys.version_info >= (3, 10),
    reason="Issue in cairo-lang. "
    "See https://github.com/starkware-libs/cairo-lang/issues/27",
)
@patch("nile.utils.debug.json.loads")
def test_debug_feedback_with_message(mock_json, caplog, args, expected):
    logging.getLogger().setLevel(logging.INFO)
    mock_json.return_value = mocked_json_message(args)

    debug(MOCK_HASH, NETWORK)

    assert expected in caplog.text



"""Command to call or invoke StarkNet smart contracts."""
import os

from dotenv import load_dotenv

from nile import accounts, deployments
from nile.core.call_or_invoke import call_or_invoke
from nile.core.deploy import deploy

try:
    from nile.signer import Signer
except ImportError:
    pass

load_dotenv()


class Account:
    """Account contract abstraction."""

    def __init__(self, signer, network):
        """Get or deploy an Account contract for the given private key."""
        self.signer = Signer(int(os.environ[signer]))
        self.network = network

        if accounts.exists(str(self.signer.public_key), network):
            signer_data = next(accounts.load(str(self.signer.public_key), network))
            self.address = signer_data["address"]
            self.index = signer_data["index"]
        else:
            address, index = self.deploy()
            self.address = address
            self.index = index

    def deploy(self):
        """Deploy an Account contract for the given private key."""
        index = accounts.current_index(self.network)
        pt = os.path.dirname(os.path.realpath(__file__)).replace("/core", "")
        overriding_path = (f"{pt}/artifacts", f"{pt}/artifacts/abis")

        address, _ = deploy(
            "Account",
            [str(self.signer.public_key)],
            self.network,
            f"account-{index}",
            overriding_path,
        )

        accounts.register(self.signer.public_key, address, index, self.network)

        return address, index

    def send(self, to, method, calldata, max_fee, nonce=None):
        """Execute a tx going through an Account contract."""
        target_address, _ = next(deployments.load(to, self.network)) or to
        calldata = [int(x) for x in calldata]

        if nonce is None:
            nonce = int(
                call_or_invoke(self.address, "call", "get_nonce", [], self.network)
            )

        if max_fee is None:
            max_fee = 0

        (call_array, calldata, sig_r, sig_s) = self.signer.sign_transaction(
            sender=self.address,
            calls=[[target_address, method, calldata]],
            nonce=nonce,
            max_fee=max_fee,
        )

        params = []
        params.append(str(len(call_array)))
        params.extend([str(elem) for sublist in call_array for elem in sublist])
        params.append(str(len(calldata)))
        params.extend([str(param) for param in calldata])
        params.append(str(nonce))

        return call_or_invoke(
            contract=self.address,
            type="invoke",
            method="__execute__",
            params=params,
            network=self.network,
            signature=[str(sig_r), str(sig_s)],
            max_fee=str(max_fee),
        )



"""nile runtime environment."""
from nile import deployments
from nile.core.account import Account
from nile.core.call_or_invoke import call_or_invoke
from nile.core.compile import compile
from nile.core.declare import declare
from nile.core.deploy import deploy
from nile.core.plugins import get_installed_plugins, skip_click_exit


class NileRuntimeEnvironment:
    """The NileRuntimeEnvironment exposes Nile functionality when running a script."""

    def __init__(self, network="localhost"):
        """Construct NRE object."""
        self.network = network
        for name, object in get_installed_plugins().items():
            setattr(self, name, skip_click_exit(object))

    def compile(self, contracts):
        """Compile a list of contracts."""
        return compile(contracts)

    def declare(self, contract, alias=None, overriding_path=None):
        """Declare a smart contract class."""
        return declare(contract, self.network, alias)

    def deploy(self, contract, arguments=None, alias=None, overriding_path=None):
        """Deploy a smart contract."""
        if arguments is None:
            arguments = []
        return deploy(contract, arguments, self.network, alias, overriding_path)

    def call(self, contract, method, params=None):
        """Call a view function in a smart contract."""
        if params is None:
            params = []
        return call_or_invoke(contract, "call", method, params, self.network)

    def invoke(self, contract, method, params=None):
        """Invoke a mutable function in a smart contract."""
        if params is None:
            params = []
        return call_or_invoke(contract, "invoke", method, params, self.network)

    def get_deployment(self, identifier):
        """Get a deployment by its identifier (address or alias)."""
        return next(deployments.load(identifier, self.network))

    def get_declaration(self, identifier):
        """Get a declared class by its identifier (class hash or alias)."""
        return next(deployments.load_class(identifier, self.network))

    def get_or_deploy_account(self, signer):
        """Get or deploy an Account contract."""
        return Account(signer, self.network)



"""Command to install a specific version of Cairo."""
import logging
import subprocess
import sys


def install():
    """Install Cairo package with the given tag."""
    logging.info("🗄  Installing Cairo")
    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", "cairo-lang", "starknet-devnet"]
    )
    logging.info("✨  Cairo successfully installed!")



"""Functions used to help debug a rejected transaction."""

import json
import logging
import os
import re
import subprocess
import time

from nile.common import (
    BUILD_DIRECTORY,
    DEPLOYMENTS_FILENAME,
    GATEWAYS,
    RETRY_AFTER_SECONDS,
)


def debug(tx_hash, network, contracts_file=None):
    """Use available contracts to help locate the error in a rejected transaction."""
    command = ["starknet", "tx_status", "--hash", tx_hash]

    if network == "mainnet":
        os.environ["STARKNET_NETWORK"] = "alpha-mainnet"
    elif network == "goerli":
        os.environ["STARKNET_NETWORK"] = "alpha-goerli"
    else:
        command.append(f"--feeder_gateway_url={GATEWAYS.get(network)}")

    logging.info(
        "⏳ Querying the network to check transaction status and identify contracts..."
    )

    while True:
        receipt = json.loads(subprocess.check_output(command))
        status = receipt["tx_status"]
        if status == "REJECTED":
            break
        output = f"Transaction status: {status}"
        if status.startswith("ACCEPTED"):
            logging.info(f"✅ {output}. No error in transaction.")
            return

        logging.info(f"🕒 {output}. Trying again in a moment...")
        time.sleep(RETRY_AFTER_SECONDS)

    error_message = receipt["tx_failure_reason"]["error_message"]
    addresses = set(
        int(address, 16)
        for address in re.findall("0x[\\da-f]{1,64}", str(error_message))
    )

    if not addresses:
        logging.warning(
            "🛑 The transaction was rejected but no contract address was identified "
            "in the error message."
        )
        logging.info(f"Error message:\n{error_message}")
        return error_message

    file = contracts_file or f"{network}.{DEPLOYMENTS_FILENAME}"
    # contracts_file should already link to compiled contracts and not ABIs
    to_contract = (lambda x: x) if contracts_file else _abi_to_build_path

    contracts = _locate_error_lines_with_abis(file, addresses, to_contract)

    if not contracts:
        logging.warning(
            "🛑 The transaction was rejected but no contract data is locally "
            "available to improve the error message."
        )
        logging.info(error_message)
        return error_message

    command += ["--contracts", ",".join(contracts), "--error_message"]
    logging.info(f"🧾 Found contracts: {contracts}")
    logging.info("⏳ Querying the network with identified contracts...")
    output = subprocess.check_output(command)

    logging.info(f"🧾 Error message:\n{output.decode()}")
    return output


def _abi_to_build_path(filename):
    return os.path.join(BUILD_DIRECTORY, os.path.basename(filename))


def _locate_error_lines_with_abis(file, addresses, to_contract):
    contracts = []
    with open(file) as file_stream:
        for line_idx, line in enumerate(file_stream):
            try:
                line_address, abi, *_ = line.split(":")
            except ValueError:
                logging.warning(
                    f"⚠ Skipping misformatted line #{line_idx+1} in {file}."
                )
                continue
            if int(line.split(":")[0], 16) in addresses:
                contracts.append(f"{line_address}:{to_contract(abi.rstrip())}")
    return contracts



"""Command to run Cairo tests written in Cairo."""
import asyncio

from nile.common import CONTRACTS_DIRECTORY, get_all_contracts


def test(contracts):
    """Compile and run all provided Cairo tests."""
    asyncio.run(_test(contracts))


async def _test(contracts):
    if len(contracts) == 0:
        print(
            f"🤖 Running all test Cairo contracts in the {CONTRACTS_DIRECTORY} directory"
        )
        for path in get_all_contracts(".test.cairo"):
            await _run_test_contract(path)
    else:
        for path in contracts:
            await _run_test_contract(path)


async def _run_test_contract(path):
    # we dinamically import starknet dependencies because
    # this module is loaded on the first nile execution,
    # potentially before installing Cairo
    from starkware.starknet.compiler.compile import compile_starknet_files
    from starkware.starknet.testing.state import StarknetState
    from starkware.starkware_utils.error_handling import StarkException

    print(f"Running tests for {path}")
    definition = compile_starknet_files([path], debug_info=True)
    starknet = await StarknetState.empty()
    contract_addr = await starknet.deploy(definition)

    for abi_entry in definition.abi:
        if abi_entry["type"] != "function":
            continue
        func_name = abi_entry["name"]
        if not func_name.startswith("test"):
            continue
        local_starknet = starknet.copy()
        try:
            await local_starknet.invoke_raw(
                contract_address=contract_addr,
                selector=func_name,
                calldata=[0 for _ in abi_entry["inputs"]],
            )
            print(f"[PASS] {func_name}")
        except StarkException as ex:
            print(f"[FAIL] {func_name}")
            print(ex.message)



"""Command to deploy StarkNet smart contracts."""
import logging

from nile import deployments
from nile.common import ABIS_DIRECTORY, BUILD_DIRECTORY, parse_information, run_command


def deploy(contract_name, arguments, network, alias, overriding_path=None):
    """Deploy StarkNet smart contracts."""
    logging.info(f"🚀 Deploying {contract_name}")
    base_path = (
        overriding_path if overriding_path else (BUILD_DIRECTORY, ABIS_DIRECTORY)
    )
    abi = f"{base_path[1]}/{contract_name}.json"

    output = run_command(contract_name, network, overriding_path, arguments=arguments)

    address, tx_hash = parse_information(output)
    logging.info(f"⏳ ️Deployment of {contract_name} successfully sent at {address}")
    logging.info(f"🧾 Transaction hash: {tx_hash}")

    deployments.register(address, abi, network, alias)
    return address, abi



"""Tests for clean command."""
from pathlib import Path
from unittest.mock import patch

import pytest

from nile.common import (
    ACCOUNTS_FILENAME,
    BUILD_DIRECTORY,
    DECLARATIONS_FILENAME,
    DEPLOYMENTS_FILENAME,
)
from nile.core.clean import clean


@pytest.fixture(autouse=True)
def tmp_working_dir(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    return tmp_path


@patch("nile.core.clean.shutil.rmtree")
@patch("nile.core.clean.os.remove")
def test_clean_already_clean(mock_os_remove, mock_shutil_rmtree):
    clean()
    mock_os_remove.assert_not_called()
    mock_shutil_rmtree.assert_not_called()


@pytest.mark.parametrize(
    "fname",
    [
        f"localhost.{ACCOUNTS_FILENAME}",
        f"localhost.{DEPLOYMENTS_FILENAME}",
        f"localhost.{DECLARATIONS_FILENAME}",
    ],
)
@patch("nile.core.clean.shutil.rmtree")
@patch("nile.core.clean.os.remove")
def test_clean_clean_files(mock_os_remove, mock_shutil_rmtree, fname):
    Path(fname).touch()
    clean()
    mock_os_remove.assert_called_once_with(fname)
    mock_shutil_rmtree.assert_not_called()


@patch("nile.core.clean.shutil.rmtree")
@patch("nile.core.clean.os.remove")
def test_clean_clean_build_dir(mock_os_remove, mock_shutil_rmtree):
    Path(BUILD_DIRECTORY).mkdir()
    clean()
    mock_os_remove.assert_not_called()
    mock_shutil_rmtree.assert_called_once_with(BUILD_DIRECTORY)



"""Command to kickstart a Nile project."""
import logging
import subprocess
import sys
from distutils.dir_util import copy_tree
from pathlib import Path

from nile.core.install import install


def init():
    """Kickstart a new Nile project."""
    # install cairo dependencies
    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", "ecdsa", "fastecdsa", "sympy"]
    )

    # install cairo within env
    install()

    # install testing dependencies
    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", "pytest", "pytest-asyncio"]
    )
    logging.info("")
    logging.info("✅ Dependencies successfully installed")

    # create project directories
    logging.info("🗄  Creating project directory tree")

    copy_tree(Path(__file__).parent.parent / "base_project", ".")

    with open("accounts.json", "w") as file:
        file.write("{}")

    logging.info("⛵️ Nile project ready! Try running:")
    logging.info("")
    logging.info("nile compile")
    logging.info("")



"""Command to call or invoke StarkNet smart contracts."""
import logging
import os
import subprocess

from nile import deployments
from nile.common import GATEWAYS


def call_or_invoke(
    contract, type, method, params, network, signature=None, max_fee=None
):
    """Call or invoke functions of StarkNet smart contracts."""
    address, abi = next(deployments.load(contract, network))

    command = [
        "starknet",
        type,
        "--address",
        address,
        "--abi",
        abi,
        "--function",
        method,
    ]

    if network == "mainnet":
        os.environ["STARKNET_NETWORK"] = "alpha-mainnet"
    elif network == "goerli":
        os.environ["STARKNET_NETWORK"] = "alpha-goerli"
    else:
        gateway_prefix = "feeder_gateway" if type == "call" else "gateway"
        command.append(f"--{gateway_prefix}_url={GATEWAYS.get(network)}")

    if len(params) > 0:
        command.append("--inputs")
        command.extend(params)

    if signature is not None:
        command.append("--signature")
        command.extend(signature)

    if max_fee is not None:
        command.append("--max_fee")
        command.append(max_fee)

    try:
        output = subprocess.check_output(command).strip().decode("utf-8")
        return output

    except subprocess.CalledProcessError:
        p = subprocess.Popen(command, stderr=subprocess.PIPE)
        _, error = p.communicate()

        if "max_fee must be bigger than 0" in error.decode():
            logging.error(
                """
                \n😰 Whoops, looks like max fee is missing. Try with:\n
                --max_fee=`MAX_FEE`
                """
            )
        else:
            raise



"""Tests for cli.py."""
import json
import shutil
import sys
from multiprocessing import Process
from os import getpid, kill
from pathlib import Path
from signal import SIGINT
from threading import Timer
from time import sleep
from unittest.mock import patch
from urllib.error import URLError
from urllib.request import urlopen

import pytest
from click.testing import CliRunner

from nile.cli import cli
from nile.common import (
    ABIS_DIRECTORY,
    BUILD_DIRECTORY,
    CONTRACTS_DIRECTORY,
    NODE_FILENAME,
)

RESOURCES_DIR = Path(__file__).parent / "resources"
MOCK_HASH = "0x123"


pytestmark = pytest.mark.end_to_end


@pytest.fixture(autouse=True)
def tmp_working_dir(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    return tmp_path


def create_process(target, args):
    """Spawns another process in Python."""
    p = Process(target=target, args=args)
    return p


def start_node(seconds, node_args):
    """Start node with host and port specified in node_args and life in seconds."""
    # Timed kill command with SIGINT to close Node process
    Timer(seconds, lambda: kill(getpid(), SIGINT)).start()
    CliRunner().invoke(cli, ["node", *node_args])


def check_node(p, seconds, gateway_url):
    """Check if node is running while spawned process is alive."""
    check_runs = 0
    while p.is_alive and check_runs < seconds:
        try:
            status = urlopen(gateway_url + "is_alive").getcode()
            return status
        except URLError:
            check_runs += 1
            sleep(1)
            continue


def test_clean():
    # The implementation is already thoroughly covered by unit tests, so here
    # we just check whether the command completes successfully.
    result = CliRunner().invoke(cli, ["clean"])
    assert result.exit_code == 0


@pytest.mark.parametrize(
    "args, expected",
    [
        ([], {"contract_1.json", "contract_2.json"}),
        (["contract_1.cairo"], {"contract_1.json"}),
        (["contract_2.cairo"], {"contract_2.json"}),
    ],
)
@pytest.mark.xfail(
    sys.version_info >= (3, 10),
    reason="Issue in cairo-lang. "
    "See https://github.com/starkware-libs/cairo-lang/issues/27",
)
def test_compile(args, expected):
    contract_source = RESOURCES_DIR / "contracts" / "contract.cairo"

    target_dir = Path(CONTRACTS_DIRECTORY)
    target_dir.mkdir()

    shutil.copyfile(contract_source, target_dir / "contract_1.cairo")
    shutil.copyfile(contract_source, target_dir / "contract_2.cairo")

    abi_dir = Path(ABIS_DIRECTORY)
    build_dir = Path(BUILD_DIRECTORY)

    assert not abi_dir.exists()
    assert not build_dir.exists()

    result = CliRunner().invoke(cli, ["compile", *args])
    assert result.exit_code == 0

    assert {f.name for f in abi_dir.glob("*.json")} == expected
    assert {f.name for f in build_dir.glob("*.json")} == expected


@pytest.mark.parametrize(
    "args, expected",
    [
        ([], "http://127.0.0.1:5050/"),
        (["--host", "localhost", "--port", "5001"], "http://localhost:5001/"),
    ],
)
@pytest.mark.xfail(
    sys.version_info >= (3, 10),
    reason="Issue in cairo-lang. "
    "See https://github.com/starkware-libs/cairo-lang/issues/27",
)
def test_node(args, expected):
    # Node life
    seconds = 15

    if args == []:
        host, port = "127.0.0.1", 5050
    else:
        host, port = args[1], int(args[3])

    if host == "127.0.0.1":
        network = "localhost"
    else:
        network = host

    gateway_url = f"http://{host}:{port}/"

    # Spawn process to start StarkNet local network with specified port
    # i.e. $ nile node --host localhost --port 5001
    p = create_process(target=start_node, args=(seconds, args))
    p.start()

    # Check node heartbeat and assert that it is running
    status = check_node(p, seconds, gateway_url)
    p.join()
    assert status == 200

    # Assert network and gateway_url is correct in node.json file
    file = NODE_FILENAME
    with open(file, "r") as f:
        gateway = json.load(f)
    assert gateway.get(network) == expected


@pytest.mark.parametrize(
    "args",
    [
        ([MOCK_HASH, "--network", "goerli"]),
        ([MOCK_HASH, "--network", "mainnet", "--contracts_file", "example.txt"]),
    ],
)
@patch("nile.utils.debug.subprocess")
def test_debug(mock_subprocess, args):
    # debug will hang without patch
    mock_subprocess.check_output.return_value = json.dumps({"tx_status": "ACCEPTED"})

    result = CliRunner().invoke(cli, ["debug", *args])

    # Check status
    assert result.exit_code == 0

    # Setup and assert expected output
    expected = ["starknet", "tx_status", "--hash", MOCK_HASH]

    mock_subprocess.check_output.assert_called_once_with(expected)



"""
Tests for signer.py.

A high-level synchronization check between the Account artifacts and Signer module.
"""

import asyncio

import pytest
from starkware.starknet.services.api.contract_class import ContractClass
from starkware.starknet.testing.starknet import Starknet

from nile.signer import Signer

SIGNER = Signer(12345678987654321)


def get_account_definition():
    with open("src/nile/artifacts/Account.json", "r") as fp:
        return ContractClass.loads(fp.read())


async def send_transaction(
    signer, account, to, selector_name, calldata, nonce=None, max_fee=0
):
    return await send_transactions(
        signer, account, [(to, selector_name, calldata)], nonce, max_fee
    )


async def send_transactions(signer, account, calls, nonce=None, max_fee=0):
    if nonce is None:
        execution_info = await account.get_nonce().call()
        (nonce,) = execution_info.result

    build_calls = []
    for call in calls:
        build_call = list(call)
        build_call[0] = hex(build_call[0])
        build_calls.append(build_call)

    (call_array, calldata, sig_r, sig_s) = signer.sign_transaction(
        hex(account.contract_address), build_calls, nonce, max_fee
    )
    return await account.__execute__(call_array, calldata, nonce).invoke(
        signature=[sig_r, sig_s]
    )


@pytest.fixture(scope="module")
def event_loop():
    return asyncio.new_event_loop()


@pytest.mark.asyncio
async def test_execute():
    starknet = await Starknet.empty()
    account = await starknet.deploy(
        contract_class=get_account_definition(),
        constructor_calldata=[SIGNER.public_key],
    )
    contract = await starknet.deploy(
        "tests/resources/contracts/contract.cairo",
    )

    # Ensures balance is zero
    execution_info = await contract.get_balance().call()
    assert execution_info.result == (0,)

    # Single tx
    await send_transaction(
        SIGNER, account, contract.contract_address, "increase_balance", [1]
    )
    execution_info = await contract.get_balance().call()
    assert execution_info.result == (1,)

    # Multicall tx
    await send_transactions(
        SIGNER,
        account,
        [
            (contract.contract_address, "increase_balance", [1]),
            (contract.contract_address, "increase_balance", [1]),
        ],
    )
    execution_info = await contract.get_balance().call()
    assert execution_info.result == (3,)



"""nile common module."""
import json
import os
import re
import subprocess

CONTRACTS_DIRECTORY = "contracts"
BUILD_DIRECTORY = "artifacts"
TEMP_DIRECTORY = ".temp"
ABIS_DIRECTORY = f"{BUILD_DIRECTORY}/abis"
DEPLOYMENTS_FILENAME = "deployments.txt"
DECLARATIONS_FILENAME = "declarations.txt"
ACCOUNTS_FILENAME = "accounts.json"
NODE_FILENAME = "node.json"
RETRY_AFTER_SECONDS = 30


def _get_gateway():
    """Get the StarkNet node details."""
    try:
        with open(NODE_FILENAME, "r") as f:
            gateway = json.load(f)
            return gateway

    except FileNotFoundError:
        with open(NODE_FILENAME, "w") as f:
            f.write('{"localhost": "http://127.0.0.1:5050/"}')


GATEWAYS = _get_gateway()


def get_all_contracts(ext=None, directory=None):
    """Get all cairo contracts in the default contract directory."""
    if ext is None:
        ext = ".cairo"

    files = list()
    for (dirpath, _, filenames) in os.walk(
        directory if directory else CONTRACTS_DIRECTORY
    ):
        files += [
            os.path.join(dirpath, file) for file in filenames if file.endswith(ext)
        ]

    return files


def run_command(
    contract_name, network, overriding_path=None, operation="deploy", arguments=None
):
    """Execute CLI command with given parameters."""
    base_path = (
        overriding_path if overriding_path else (BUILD_DIRECTORY, ABIS_DIRECTORY)
    )
    contract = f"{base_path[0]}/{contract_name}.json"
    command = ["starknet", operation, "--contract", contract]

    if arguments:
        command.append("--inputs")
        command.extend([argument for argument in arguments])

    if network == "mainnet":
        os.environ["STARKNET_NETWORK"] = "alpha-mainnet"
    elif network == "goerli":
        os.environ["STARKNET_NETWORK"] = "alpha-goerli"
    else:
        command.append(f"--gateway_url={GATEWAYS.get(network)}")

    return subprocess.check_output(command)


def parse_information(x):
    """Extract information from deploy/declare command."""
    # address is 64, tx_hash is 64 chars long
    address, tx_hash = re.findall("0x[\\da-f]{1,64}", str(x))
    return address, tx_hash



"""
Tests for plugins in core module.

Only unit tests for now.
"""

from unittest.mock import patch

import click

from nile.core.plugins import get_installed_plugins, load_plugins, skip_click_exit


def test_skip_click_exit():
    @click.command()
    @click.argument("a", type=int)
    @click.argument("b", type=int)
    def dummy_method(a, b):
        return a + b

    decorated = skip_click_exit(dummy_method)
    decorated_result = decorated(["1", "2"])

    assert callable(decorated)
    assert decorated_result == 3


def testget_installed_plugins():
    class Dummy:
        value = "nile.core.plugins.get_installed_plugins"
        name = "get_installed_plugins"

    with patch("nile.core.plugins.entry_points", return_value=[Dummy()]):
        installed_plugins = get_installed_plugins()
        assert "get_installed_plugins" in installed_plugins


def test_load_plugins():
    @click.group()
    def cli():
        """Nile CLI group."""
        pass

    @click.command()
    def dummy():
        print("dummy_result")

    with patch(
        "nile.core.plugins.get_installed_plugins", return_value={"dummy": dummy}
    ):
        app = load_plugins(cli)
        assert callable(app)



"""Command to start StarkNet local network."""
import json
import subprocess

from nile.common import NODE_FILENAME


def node(host="127.0.0.1", port=5050):
    """Start StarkNet local network."""
    try:
        # Save host and port information to be used by other commands
        file = NODE_FILENAME
        if host == "127.0.0.1":
            network = "localhost"
        else:
            network = host
        gateway_url = f"http://{host}:{port}/"
        gateway = {network: gateway_url}

        with open(file, "w+") as f:
            json.dump(gateway, f)

        # Start network
        subprocess.check_call(["starknet-devnet", "--host", host, "--port", str(port)])

    except FileNotFoundError:
        print("")
        print("😰 Could not find starknet-devnet, is it installed? Try with:\n")
        print("   pip install starknet-devnet")
        print("")



"""Nile plugins utility methods."""
import functools
import importlib
from typing import Dict

from click import Group
from importlib_metadata import entry_points


def skip_click_exit(func):
    """
    When a click command is executed programatically a SystemExit is executed.

    This decorator avoids SystemExit when execution is finished.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # add standalone_mode=False to command execution
        # this enables command returns
        # more info: https://click.palletsprojects.com/en/5.x/commands/?highlight=standalone_mode#command-return-values # noqa: E501
        try:
            return func(*args, standalone_mode=False, **kwargs)
        # click commands always raise a SystemExit
        # this avoid exiting the command execution in NRE
        except SystemExit:
            pass

    return wrapper


def get_installed_plugins() -> Dict:
    """
    Get the name and object of installed Nile plugins.

    Returns:
        Dict: dictionary with plugin name as key and value with the plugin object
    """
    discovered_plugins = entry_points(group="nile_plugins")
    loaded_plugins = {}
    for plugin in discovered_plugins:
        package_name = plugin.value.rsplit(".", 1)[0]
        package_object = importlib.import_module(package_name)
        plugin_call = getattr(package_object, plugin.name)
        loaded_plugins[plugin.name] = plugin_call
    return loaded_plugins


def load_plugins(app) -> Group:
    """
    Add plugin entrypoints to Click app.

    Args:
        app (click.group): main cli app

    Returns:
        Group: click app with plugins commands added
    """
    for name, plugin_object in get_installed_plugins().items():
        app.add_command(plugin_object, name)
    return app



#!/usr/bin/env python
"""Nile CLI entry point."""
import logging

import click

from nile.core.account import Account
from nile.core.call_or_invoke import call_or_invoke as call_or_invoke_command
from nile.core.clean import clean as clean_command
from nile.core.compile import compile as compile_command
from nile.core.declare import declare as declare_command
from nile.core.deploy import deploy as deploy_command
from nile.core.init import init as init_command
from nile.core.install import install as install_command
from nile.core.node import node as node_command
from nile.core.plugins import load_plugins
from nile.core.run import run as run_command
from nile.core.test import test as test_command
from nile.core.version import version as version_command
from nile.utils.debug import debug as debug_command

logging.basicConfig(level=logging.DEBUG, format="%(message)s")

NETWORKS = ("localhost", "goerli", "mainnet")


def network_option(f):
    """Configure NETWORK option for the cli."""
    return click.option(  # noqa: E731
        "--network",
        envvar="STARKNET_NETWORK",
        default="localhost",
        help=f"Select network, one of {NETWORKS}",
        callback=_validate_network,
    )(f)


def _validate_network(_ctx, _param, value):
    """Normalize network values."""
    # normalize goerli
    if "goerli" in value or "testnet" in value:
        return "goerli"
    # normalize localhost
    if "localhost" in value or "127.0.0.1" in value:
        return "localhost"
    # check if value is accepted
    if value in NETWORKS:
        return value
    # raise if value is invalid
    raise click.BadParameter(f"'{value}'. Use one of {NETWORKS}")


@click.group()
def cli():
    """Nile CLI group."""
    pass


@cli.command()
def init():
    """Nile CLI group."""
    init_command()


@cli.command()
def install():
    """Install Cairo."""
    install_command()


@cli.command()
@click.argument("path", nargs=1)
@network_option
def run(path, network):
    """Run Nile scripts with NileRuntimeEnvironment."""
    run_command(path, network)


@cli.command()
@click.argument("artifact", nargs=1)
@click.argument("arguments", nargs=-1)
@network_option
@click.option("--alias")
def deploy(artifact, arguments, network, alias):
    """Deploy StarkNet smart contract."""
    deploy_command(artifact, arguments, network, alias)


@cli.command()
@click.argument("artifact", nargs=1)
@network_option
@click.option("--alias")
def declare(artifact, network, alias):
    """Declare StarkNet smart contract."""
    declare_command(artifact, network, alias)


@cli.command()
@click.argument("signer", nargs=1)
@network_option
def setup(signer, network):
    """Set up an Account contract."""
    Account(signer, network)


@cli.command()
@click.argument("signer", nargs=1)
@click.argument("contract_name", nargs=1)
@click.argument("method", nargs=1)
@click.argument("params", nargs=-1)
@click.option("--max_fee", nargs=1)
@network_option
def send(signer, contract_name, method, params, network, max_fee=None):
    """Invoke a contract's method through an Account. Same usage as nile invoke."""
    account = Account(signer, network)
    print(
        "Calling {} on {} with params: {}".format(
            method, contract_name, [x for x in params]
        )
    )
    out = account.send(contract_name, method, params, max_fee=max_fee)
    print(out)


@cli.command()
@click.argument("contract_name", nargs=1)
@click.argument("method", nargs=1)
@click.argument("params", nargs=-1)
@click.option("--max_fee", nargs=1)
@network_option
def invoke(contract_name, method, params, network, max_fee=None):
    """Invoke functions of StarkNet smart contracts."""
    out = call_or_invoke_command(
        contract_name, "invoke", method, params, network, max_fee=max_fee
    )
    print(out)


@cli.command()
@click.argument("contract_name", nargs=1)
@click.argument("method", nargs=1)
@click.argument("params", nargs=-1)
@network_option
def call(contract_name, method, params, network):
    """Call functions of StarkNet smart contracts."""
    out = call_or_invoke_command(contract_name, "call", method, params, network)
    print(out)


@cli.command()
@click.argument("contracts", nargs=-1)
def test(contracts):
    """
    Run cairo test contracts.

    $ nile test
      Compiles all test contracts in CONTRACTS_DIRECTORY

    $ nile test contracts/MyContract.test.cairo
      Runs tests in MyContract.test.cairo

    $ nile test contracts/foo.test.cairo contracts/bar.test.cairo
      Runs tests in foo.test.cairo and bar.test.cairo
    """
    test_command(contracts)


@cli.command()
@click.argument("contracts", nargs=-1)
@click.option("--directory")
@click.option("--account_contract", is_flag="True")
@click.option("--disable-hint-validation", is_flag=True)
def compile(contracts, directory, account_contract, disable_hint_validation):
    """
    Compile cairo contracts.

    $ compile.py
      Compiles all contracts in CONTRACTS_DIRECTORY

    $ compile.py contracts/MyContract.cairo
      Compiles MyContract.cairo

    $ compile.py contracts/foo.cairo contracts/bar.cairo
      Compiles foo.cairo and bar.cairo
    """
    compile_command(contracts, directory, account_contract, disable_hint_validation)


@cli.command()
def clean():
    """Remove default build directory."""
    clean_command()


@cli.command()
@click.option("--host", default="127.0.0.1")
@click.option("--port", default=5050)
def node(host, port):
    """Start StarkNet local network.

    $ nile node
      Start StarkNet local network at port 5050

    $ nile node --host HOST --port 5001
      Start StarkNet network on address HOST listening at port 5001
    """
    node_command(host, port)


@cli.command()
@click.version_option()
def version():
    """Print out toolchain version."""
    version_command()


@cli.command()
@click.argument("tx_hash", nargs=1)
@network_option
@click.option("--contracts_file", nargs=1)
def debug(tx_hash, network, contracts_file):
    """Locate an error in a transaction using contracts."""
    debug_command(tx_hash, network, contracts_file)


cli = load_plugins(cli)


if __name__ == "__main__":
    cli()



"""Command to compile cairo files."""
import logging
import os
import subprocess

from nile.common import (
    ABIS_DIRECTORY,
    BUILD_DIRECTORY,
    CONTRACTS_DIRECTORY,
    get_all_contracts,
)


def compile(
    contracts, directory=None, account_contract=False, disable_hint_validation=False
):
    """Compile cairo contracts to default output directory."""
    # to do: automatically support subdirectories

    contracts_directory = directory if directory else CONTRACTS_DIRECTORY

    if not os.path.exists(ABIS_DIRECTORY):
        logging.info(f"📁 Creating {ABIS_DIRECTORY} to store compilation artifacts")
        os.makedirs(ABIS_DIRECTORY, exist_ok=True)

    all_contracts = contracts

    if len(contracts) == 0:
        logging.info(
            f"🤖 Compiling all Cairo contracts in the {contracts_directory} directory"
        )
        all_contracts = get_all_contracts(directory=contracts_directory)

    results = [
        _compile_contract(
            contract, contracts_directory, account_contract, disable_hint_validation
        )
        for contract in all_contracts
    ]
    failed_contracts = [c for (c, r) in zip(all_contracts, results) if r != 0]
    failures = len(failed_contracts)

    if failures == 0:
        logging.info("✅ Done")
    else:
        exp = f"{failures} contract"
        if failures > 1:
            exp += "s"  # pluralize
        logging.info(f"🛑 Failed to compile the following {exp}:")
        for contract in failed_contracts:
            logging.info(f"   {contract}")


def _compile_contract(
    path, directory=None, account_contract=False, disable_hint_validation=False
):
    base = os.path.basename(path)
    filename = os.path.splitext(base)[0]
    logging.info(f"🔨 Compiling {path}")
    contracts_directory = directory if directory else CONTRACTS_DIRECTORY

    cmd = f"""
    starknet-compile {path} \
        --cairo_path={contracts_directory}
        --output {BUILD_DIRECTORY}/{filename}.json \
        --abi {ABIS_DIRECTORY}/{filename}.json
    """

    if account_contract or filename.endswith("Account"):
        cmd = cmd + "--account_contract"

    if disable_hint_validation:
        cmd = cmd + "--disable_hint_validation"

    process = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
    process.communicate()
    return process.returncode



"""Command to clean artifacts from workspace."""
import logging
import os
import shutil

from nile.common import (
    ACCOUNTS_FILENAME,
    BUILD_DIRECTORY,
    DECLARATIONS_FILENAME,
    DEPLOYMENTS_FILENAME,
)


def clean():
    """Remove artifacts from workspace."""
    local_files = [
        f"localhost.{DEPLOYMENTS_FILENAME}",
        f"localhost.{DECLARATIONS_FILENAME}",
        f"localhost.{ACCOUNTS_FILENAME}",
        BUILD_DIRECTORY,
    ]

    for file in local_files:
        if os.path.exists(file):
            logging.info(f"🚮 Deleting {file}")

            if file is BUILD_DIRECTORY:
                shutil.rmtree(BUILD_DIRECTORY)
            else:
                os.remove(file)

    logging.info("✨ Workspace clean, keep going!")



"""Command to run Nile scripts."""
import logging
from importlib.machinery import SourceFileLoader

from nile.nre import NileRuntimeEnvironment


def run(path, network):
    """Run nile scripts passing on the NRE object."""
    logger = logging.getLogger()
    logger.disabled = True
    script = SourceFileLoader("script", path).load_module()
    nre = NileRuntimeEnvironment(network)
    script.run(nre)



"""
Tests for nre module.

Only unit tests for now.
"""

from unittest.mock import patch

import click

from nile.nre import NileRuntimeEnvironment


def test_nre_loaded_plugins():
    @click.command()
    def dummy():
        print("dummy_result")

    @click.command()
    @click.argument("a", type=int)
    @click.argument("b", type=int)
    def dummy_params(a, b):
        return a + b

    with patch(
        "nile.nre.get_installed_plugins",
        return_value={"dummy": dummy, "dummy_params": dummy_params},
    ):
        nre = NileRuntimeEnvironment()
        assert callable(nre.dummy)

        nre_result = nre.dummy_params(["1", "2"])
        assert 3 == nre_result



"""contract.cairo test file."""
import os

import pytest
from starkware.starknet.testing.starknet import Starknet

# The path to the contract source code.
CONTRACT_FILE = os.path.join("contracts", "contract.cairo")


# The testing library uses python's asyncio. So the following
# decorator and the ``async`` keyword are needed.
@pytest.mark.asyncio
async def test_increase_balance():
    """Test increase_balance method."""
    # Create a new Starknet class that simulates the StarkNet
    # system.
    starknet = await Starknet.empty()

    # Deploy the contract.
    contract = await starknet.deploy(
        source=CONTRACT_FILE,
    )

    # Invoke increase_balance() twice.
    await contract.increase_balance(amount=10).invoke()
    await contract.increase_balance(amount=20).invoke()

    # Check the result of get_balance().
    execution_info = await contract.get_balance().call()
    assert execution_info.result == (30,)



"""Tests for declare command."""
import logging
from unittest.mock import patch

import pytest

from nile.common import DECLARATIONS_FILENAME
from nile.core.declare import alias_exists, declare


@pytest.fixture(autouse=True)
def tmp_working_dir(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    return tmp_path


CONTRACT = "contract"
NETWORK = "goerli"
ALIAS = "alias"
PATH = "path"
RUN_OUTPUT = b"output"
HASH = 111
TX_HASH = 222


def test_alias_exists():
    # when alias does not exist
    assert alias_exists(ALIAS, NETWORK) is False

    # when alias exists
    with patch("nile.core.declare.deployments.load_class") as mock_load:
        mock_load.__iter__.side_effect = HASH
        assert alias_exists(ALIAS, NETWORK) is True


@pytest.mark.parametrize(
    "args, exp_command, exp_register",
    [
        (
            [CONTRACT, NETWORK],  # args
            [CONTRACT, NETWORK, None],  # expected command
            [HASH, NETWORK, None],  # expected register
        ),
        (
            [CONTRACT, NETWORK, ALIAS],  # args
            [CONTRACT, NETWORK, None],  # expected command
            [HASH, NETWORK, ALIAS],  # expected register
        ),
        (
            [CONTRACT, NETWORK, ALIAS, PATH],  # args
            [CONTRACT, NETWORK, PATH],  # expected command
            [HASH, NETWORK, ALIAS],  # expected register
        ),
    ],
)
@patch("nile.core.declare.run_command", return_value=RUN_OUTPUT)
@patch("nile.core.declare.parse_information", return_value=[HASH, TX_HASH])
@patch("nile.core.declare.deployments.register_class_hash")
def test_declare(
    mock_register, mock_parse, mock_run_cmd, caplog, args, exp_command, exp_register
):
    logging.getLogger().setLevel(logging.INFO)

    # check return value
    res = declare(*args)
    assert res == HASH

    # check internals
    mock_run_cmd.assert_called_once_with(*exp_command, operation="declare")
    mock_parse.assert_called_once_with(RUN_OUTPUT)
    mock_register.assert_called_once_with(*exp_register)

    # check logs
    assert f"🚀 Declaring {CONTRACT}" in caplog.text
    assert f"⏳ Declaration of {CONTRACT} successfully sent at {HASH}" in caplog.text
    assert f"🧾 Transaction hash: {TX_HASH}" in caplog.text


@patch("nile.core.declare.alias_exists", return_value=True)
def test_declare_duplicate_hash(mock_alias_check):

    with pytest.raises(Exception) as err:
        declare(ALIAS, NETWORK)

        assert (
            f"Alias {ALIAS} already exists in {NETWORK}.{DECLARATIONS_FILENAME}"
            in str(err.value)
        )
